#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_LIBVNCSERVER
#include <rfb/rfb.h>
#include <rfb/keysym.h>
#endif /* HAVE_LIBVNCSERVER */

#include "dcpu16.h"
#include "hw_keyboard.h"

#define BUF_SZ 32

static dcpu16_hw_signal_t keyboard_reset_;
static dcpu16_hw_signal_t keyboard_cycle_;
static dcpu16_hw_signal_t keyboard_hwi_;
static struct dcpu16_hw hw_ = {
    .vm     = NULL,
    .name_  = "Generic Keyboard (compatible)",
    .id_l   = 0x7406,
    .id_h   = 0x30cf,
    .ver    = 0x0001,
    .mfg_l  = 0x0000,
    .mfg_h  = 0x0000,
    .hwi    = keyboard_hwi_,
    .cycle  = keyboard_cycle_,
    .reset  = keyboard_reset_,
    .data   = (struct keyboard_ *)NULL
};

struct keyboard_ {
    char *buf;
    size_t buf_sz;
    size_t buf_head;
    size_t buf_tail;
    DCPU16_WORD interrupt_message;
    unsigned char keys_pressed[256];
};

#ifdef HAVE_LIBVNCSERVER

static const struct keysym_map_ {
    unsigned int rfb_start;
    unsigned int rfb_end;
    DCPU16_WORD dcpu_start;
} keymap_[] = {
    { XK_space,     XK_asciitilde,  0x20 }, /* ASCII range */
    { XK_Delete,    XK_Delete,      0x7f }, /* ASCII del */
    { XK_BackSpace, XK_BackSpace,   0x10 }, /* bs */
    { XK_Return,    XK_Return,      0x11 }, /* ret */
    { XK_Insert,    XK_Insert,      0x12 }, /* ins */
    { XK_Delete,    XK_Delete,      0x13 }, /* del */
    { XK_Up,        XK_Up,          0x80 }, /* arrow up */
    { XK_Down,      XK_Down,        0x81 }, /* arrow down */
    { XK_Left,      XK_Left,        0x82 }, /* arrow left */
    { XK_Right,     XK_Right,       0x83 }, /* arrow right */
    { XK_Shift_L,   XK_Shift_R,     0x90 }, /* shift range */
    { XK_Control_L, XK_Control_R,   0x91 }, /* control range */
    { 0, 0, 0x0 }
};

static int keysym_rfbtodcpu(unsigned int rfb, DCPU16_WORD *dcpu) {
    const struct keysym_map_ *map;

    for (map = keymap_; map->rfb_start; map++) {
        if (rfb >= map->rfb_start
        &&  rfb <= map->rfb_end) {
            *dcpu = (map->dcpu_start + (rfb - map->rfb_start));
            return 0;
        }
    }

    *dcpu = 0x00;
    return -1;
}

static
void keyboard_rfbevent_(rfbBool down, rfbKeySym key, rfbClientPtr cl) {
    DCPU16_WORD dcpu_key;
    struct dcpu16_hw *hw = (struct dcpu16_hw *)cl->screen->screenData;
    struct keyboard_ *keyboard = (struct keyboard_ *)hw->data;

    keysym_rfbtodcpu(key, &dcpu_key);

    fprintf(stderr, "%s: down:%u key:0x%04x dcpu_key:0x%04x\n", __func__,
            down, key, dcpu_key);

    fprintf(stderr, "%s: hw:%p name:%s vm:%p\n", __func__, hw, hw->name_, hw->vm);

    keyboard->keys_pressed[dcpu_key] = (down ? 1 : 0);
    if (down) {
        if ((keyboard->buf_tail + 1) % keyboard->buf_sz == keyboard->buf_head) {
            hw->vm->warn_cb_("keyboard buffer overflow");
            return;
        }
        keyboard->buf[keyboard->buf_tail] = dcpu_key;
        keyboard->buf_tail += 1;
        keyboard->buf_tail %= keyboard->buf_sz;
    }
    if (keyboard->interrupt_message) {
        dcpu16_interrupt(hw->vm, keyboard->interrupt_message);
    }
}

void keyboard_vnc_associate(struct dcpu16_hw *hw, rfbScreenInfoPtr rfbScreen) {
    rfbScreen->screenData = hw;
    rfbScreen->kbdAddEvent = keyboard_rfbevent_;
}
#endif /* HAVE_LIBVNCSERVER */

static
void keyboard_reset_(struct dcpu16 *vm, void *data) {
    struct keyboard_ *keyboard = (struct keyboard_ *)data;

    (void)vm;

    keyboard->interrupt_message = 0;
    memset(keyboard->buf, 0, keyboard->buf_sz);
}

static
void keyboard_cycle_(struct dcpu16 *vm, void *data) {
    struct keyboard_ *keyboard = (struct keyboard_ *)data;

    (void)vm, (void)keyboard;
}

static
void keyboard_hwi_(struct dcpu16 *vm, void *data) {
    struct keyboard_ *keyboard = (struct keyboard_ *)data;
    DCPU16_WORD reg_a = vm->reg[DCPU16_REG_A];
    DCPU16_WORD reg_b = vm->reg[DCPU16_REG_B];

    switch (reg_a) {
        case 0: /* clear keyboard buffer */
            memset(keyboard->buf, 0, keyboard->buf_sz);
            keyboard->buf_head = 0;
            keyboard->buf_tail = 0;
            break;

        case 1: /* get next key from buffer as C */
            if (keyboard->buf_head == keyboard->buf_tail)
                break;
            vm->reg[DCPU16_REG_C] = keyboard->buf[keyboard->buf_head];
            keyboard->buf_head += 1;
            keyboard->buf_head %= keyboard->buf_sz;
            break;

        case 2: /* get currently-pressed-state of key in B as C */
            vm->reg[DCPU16_REG_C] = keyboard->keys_pressed[reg_b & 0x00ff];
            break;

        case 3: /* set interrupt state */
            keyboard->interrupt_message = reg_b;
            break;

    }
}

struct dcpu16_hw *keyboard_new(struct dcpu16 *vm) {
    struct dcpu16_hw *hw;
    char *b;

    hw = calloc(1, sizeof *hw);
    if (hw == NULL) {
        vm->warn_cb_("%s():%s", "calloc", strerror(errno));
        return NULL;
    }
    memcpy(hw, &hw_, sizeof *hw);
    hw->data = calloc(1, sizeof hw->data);
    if (hw->data == NULL) {
        vm->warn_cb_("%s():%s", "calloc", strerror(errno));
        free(hw);
        return NULL;
    }

    b = calloc(BUF_SZ, sizeof *b);
    if (b == NULL) {
        vm->warn_cb_("%s():%s", "calloc", strerror(errno));
        free(hw->data);
        free(hw);
        return NULL;
    }
    ((struct keyboard_ *)(hw->data))->buf = b;
    ((struct keyboard_ *)(hw->data))->buf_sz = BUF_SZ;

    hw->vm = vm;

    return hw;
}

void keyboard_del(struct dcpu16_hw **hw) {
    if (hw) {
        free(((struct keyboard_ *)((*hw)->data))->buf);
        ((struct keyboard_ *)((*hw)->data))->buf = NULL;

        free((*hw)->data);
        (*hw)->data = NULL;

        free(*hw);
        *hw = NULL;
    }
}
